/*
  A program to convert log files created by Destinator on Acer n35 Pocket PC

  Copyright (C) 2006 Jonas Svensson, jonass@lysator.liu.se

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA

*/

/*
  Compile on linux x86:
  gcc -Wall -lm -o d2gpx d2gpx.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <strings.h>

#define BLOCKSIZE 152
unsigned char inbuf[BLOCKSIZE];

#define FORMAT_GPX    1
#define FORMAT_CSV    2

struct {
  double lon;
  double lat;
  double alt;
  unsigned char sats;
  unsigned char fix;
  unsigned long date; // 6 digits decimal: DDMMYY
  float time;  // HHMMSSsss.
} outdata;

void print_gpx_header() {
  printf("<?xml version=\"1.0\"?>\n");
  printf("<gpx\n");
  printf("  version=\"1.0\"\n");
  printf("  creator=\"D2GPX 1.0\"\n");
  printf("  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
  printf("  xmlns=\"http://www.topografix.com/GPX/1/0\"\n");
  printf("  xsi:schemaLocation=\"http://www.topografix.com/GPX/1/0 http://www.topografix.com/GPX/1/0/gpx.xsd\">\n");
  printf("<trk>\n");
  printf("<name>ACTIVE LOG</name>\n");
  printf("<trkseg>\n");
}

void print_gpx_footer() {
  printf("</trkseg>\n");
  printf("</trk>\n");
  printf("</gpx>\n");
}

void print_header(int format) {
  switch(format) {
  case FORMAT_GPX : print_gpx_header(); break;
  }
}

void print_footer(int format) {
  switch(format) {
  case FORMAT_GPX : print_gpx_footer(); break;
  }
}

void print_csv() {
  char time[20];

  if(outdata.fix < 3)
    return;

  sprintf(time, "%010.10g", outdata.time);
  printf("%06ld, %c%c:%c%c:%c%c, ", outdata.date,
	 time[1],
	 time[2],
	 time[3],
	 time[4],
	 time[5],
	 time[6]);
  printf("%16.15g, %16.15g, ", outdata.lat, outdata.lon);
  printf("%16.15g\n", outdata.alt);
}

void print_gpx() {
  char date[7], time[20];

  if(outdata.fix < 3)
    return;

  printf("<trkpt lat=\"%16.15g\" lon=\"%16.15g\">\n", outdata.lat, outdata.lon);
  printf("<ele>%16.15g</ele>\n", outdata.alt);
  sprintf(date, "%06ld", outdata.date);
  sprintf(time, "%010.10g", outdata.time);
  printf("<time>20%c%c-%c%c-%c%cT%c%c:%c%c:%c%cZ</time>", 
	 date[4],
	 date[5],
	 date[2],
	 date[3],
	 date[0],
	 date[1],
	 time[1],
	 time[2],
	 time[3],
	 time[4],
	 time[5],
	 time[6]);

  printf("</trkpt>\n");
}

void processblock(int format){
  /* retrieve data */
  outdata.lon = *((double *)inbuf);
  outdata.lat = *((double *)(inbuf+8));
  outdata.alt = *((double *)(inbuf+16));
  outdata.date = *((long *)(inbuf+104));
  outdata.time = *((float *)(inbuf+108));
  outdata.sats = *((unsigned char *)(inbuf+52));
  outdata.fix = *((unsigned char *)(inbuf+48));
  
  /* output data */
  switch(format) {
  case FORMAT_GPX    : print_gpx(); break;
  case FORMAT_CSV    : print_csv(); break;
  default : printf("FATAL ERROR!"); exit(1);
  }
}


void process_file(int format) {
  int readlength;

  print_header(format);
  
  readlength=read(STDIN_FILENO, inbuf, BLOCKSIZE);
  while(readlength>0) {
    if(readlength == BLOCKSIZE) {
      processblock(format);
    }
    readlength=read(STDIN_FILENO, inbuf, BLOCKSIZE);
  }
  
  print_footer(format);
}

void print_help() {
  printf("D2GPX version 1.0 (c) Jonas Svensson\n");
  printf("A utility to convert logfile from Destinator 3\n");
  printf("to GPX or CSV.\n");
  printf("Usage:\n");
  printf("d2gpx gpx <file.dat >file.gpx\n");
  printf("d2gpx csv <file.dat >file.gpx\n");
}

int main(int argc, char **argv) {
  if(argc!=2) {
    print_help();
  } else if (strncasecmp(argv[1], "gpx", 3) == 0) {
    process_file(FORMAT_GPX);
  } else if (strncasecmp(argv[1], "csv", 3) == 0) {
    process_file(FORMAT_CSV);
  } else {
    print_help();
  }
  return 0;
}
